from enum import IntEnum


class SHBDOTagID(IntEnum):
    """ List of Tag ID numbers for BDO 2016 and 2019 formats.
    Must be synchronized with tags in sh_bdo.h and sh_detect.h in SH12A.
    """

    # Hex values are used for better recognition in binary files, should they be inspected by humans.
    # Group 0x0000 - 0x00FF : Miscellaneous info
    shversion = 0x00  # [char*] full version string of SHIELD-HIT12A
    shbuilddate = 0x01  # [char*] date of build
    filedate = 0x02  # [char*] bdo file creation date, RFC 2822 compliant
    user = 0x03  # [char *] optional login name
    host = 0x04  # [char *] optional host where this file was created
    format = 0x05  # [int]   optional ID describing which flavour the format is, to help .bdo parsers

    # Group 0xAA00 - 0xAAFF : Runtime variables
    rt_nstat = 0xAA00  # number of actually simulated particles
    rt_time = 0xAA01  # [unsigned long int] optional runtime in seconds
    rt_timesim = 0xAA02  # optional simulation time in seconds, excluding initialization.

    # Group 0xCB00 - 0xCBFF : Beam configuration
    jpart0 = 0xCB00  # [int] primary particle ID, in SH12A JPART terminology (32768 = INVALID)
    apro0 = 0xCB01  # [int] number of nucleons A of the projectile - only written if nucleons > 0
    zpro0 = 0xCB02  # [int] charge Z of the projectile, may also be negative (32768 = INVALID)
    beamx = 0xCB03  # [float] start position of the beam - X coordinate
    beamy = 0xCB04  # [float] start position of the beam - Y coordinate
    beamz = 0xCB05  # [float] start position of the beam - Z coordinate
    sigmax = 0xCB06  # [float] lateral extension of the beam in X direction
    sigmay = 0xCB07  # [float] lateral extension of the beam in Y direction
    tmax0 = 0xCB08  # [float] initial projectile energy (unit depends on projectile type)
    sigmat0 = 0xCB09  # [float] energy spread of the primary particle
    beamtheta = 0xCB0A  # [float] polar angle
    beamphi = 0xCB0B  # [float] azimuth angle
    beamdivx = 0xCB0C  # [float] beam divergence - X coordinate
    beamdivy = 0xCB0D  # [float] beam divergence - Y coordinate
    beamdivk = 0xCB0E  # [float] beam divergence - focus
    tmax0mev = 0xCB0F  # [double] initial projectile energy, always in [MeV]
    tmax0amu = 0xCB10  # [double] initial projectile energy in [MeV/amu] - only written if mass > 1e-6 u
    tmax0nuc = 0xCB11  # [double] initial projectile energy in [MeV/nucl] - only written if nucleons > 0

    # Group 0xCC00 - 0xCCFF : Configuration
    dele = 0xCC00
    demin = 0xCC01
    itypst = 0xCC02
    itypms = 0xCC03
    oln = 0xCC04
    inucre = 0xCC05
    iemtrans = 0xCC06
    iextspec = 0xCC07
    intrfast = 0xCC08
    intrslow = 0xCC09
    apzlscl = 0xCC0A
    ioffset = 0xCC0B
    irifimc = 0xCC0C
    irifitrans = 0xCC0D
    irifizone = 0xCC0E
    ext_nproj = 0xCC0F
    ext_ptvdose = 0xCC10
    ixfirs = 0xCC11

    # Group 0xCE00 - 0xCEFF : CT specific tags
    ct_ang = 0xCE00  # holds two doubles with the couch and gantry angle
    ct_icnt = 0xCE01  # holds three
    ct_len = 0xCE02  # holds three

    # Group 0xDD00 - 0xDD11 : old detect.f fortran Detector / page specific tags. See sh_detect.h
    det_geotyp = 0xDD00  # may differ from est_geotyp in case of spc
    det_nbin = 0xDD01  # idet(1-3) (len=3) number of bins x,y,z
    det_part = 0xDD02  # idet(4) particle type which was scored
    det_dtype = 0xDD03  # idet(5) detector type
    det_partz = 0xDD04  # idet(6)
    det_parta = 0xDD05  # idet(7)
    det_dmat = 0xDD06  # idet(8)
    det_nbine = 0xDD07  # idet(9) number of bins in diff scorer, negative means log binning
    det_difftype = 0xDD08  # idet(10) detector type for differential scorer (i.e. angle, energy, let)
    det_zonestart = 0xDD09  # idet(11)
    det_dsize = 0xDD0A  # idet(12)
    det_dsizexyz = 0xDD0B  # idet(13)

    det_xyz_start = 0xDD0C  # det(1-3)
    det_xyz_stop = 0xDD0D  # det(4-6)
    det_dif_start = 0xDD0E  # det(7)
    det_dif_stop = 0xDD0F  # det(8)
    det_voxvol = 0xDD10  # det(9)
    det_thresh = 0xDD11  # det(10) lower energy scoring threshold

    # Page: meta-data */
    # Group 0xDD30 - 0xDDFF : page specific tags. */
    detector_type = 0xDD30  # /* detector_type, this starts new page */
    page_number = 0xDD31  # /* Number of this detector */
    page_normalized = 0xDD32  # Flags for page->postproc on how to postprocess the data in SHBDO_PAG_DATA
    """
    Given:
        - the data in page->data as x_j
        - for j instances of this simulation
        - which was done with I_j number of paritcles.

        The resulting data will be termed X and has the units given by SHBDO_PAG_DATA_UNIT

        0: X = x_1                                for GEOMAP type scorers
        1: X = sum_j x_j                          COUNT, ...
        2: X = (sum_j x_j) / (sum_j I_j)          NORMCOUNT, ...
        3: X = (sum_j x_j * I_j) / (sum_j I_j)    LET, ...
    """

    page_scale_factor = 0xDD33  # /* if set and != 1.0 the data set was multiplied with this factor */
    page_offset = 0xDD34  # /* if set and != 0.0 the data set was offset with this value */
    page_medium_transport = 0xDD35  # /* [future] ASCII-string for detector medium set in geo.dat */
    page_medium_scoring = 0xDD36  # /* [future] ASCII-string for detector medium set in detect.dat scoring */
    page_unit_ids = 0xDD37  # /* Unit IDs according to sh_units.h. Set for detector and the two diff. bins */

    # /* page data */
    data_block = 0xDDBB  # /* data block, identical to SHBDO_DET_DATA */
    detector_unit = 0xDDBC  # /* ASCII string unit, including any differentials */

    # /* Page differential data */
    page_diff_flag = 0xDDD0  # /* flags if 1 or 2 differential binning was set. 1 for set, -1 for set as log10. */
    page_diff_type = 0xDDD1  # /* array holding 1 or 2 number of bins, for type of differential */
    page_diff_start = 0xDDD2  # /* array holding 1 or 2 lower bounds, for 1-D or 2-D respectively */
    page_diff_stop = 0xDDD3  # /* array holding 1 or 2 upper bounds, for 1-D or 2-D respectively */
    page_diff_size = 0xDDD4  # /* array holding 1 or 2 number of bins, for 1-D or 2-D respectively */
    page_diff_units = 0xDDD5  # /* ASCII string of ;-separated units along each dimension. %s;%s;%s where latter
    #          two %s are the differential units, and the first is the data in
    #          non-differential form. */

    # /* Filter data attached to page */
    page_filter_name = 0xDDF0  # /* name of filter containing one or more rules */
    page_filter_rules_no = 0xDDF1  # /* number of filter rules applied */
    page_filter_e_min = 0xDDF2  # /* lower energy threshold, emin */
    page_filter_emax = 0xDDF3  # /* upper energy threshold, emin */

    # Group 0xEE00 - 0xEEFF : Estimator specific tags
    # Geometry, as in gE0metry
    geometry_type = 0xE000  # geometry type ID, see SH_SGEO_* in sh_scoredef.h
    geometry_name = 0xE001  # /* User-given name of this geometry */
    geo_p_start = 0xE002  # /* start values, e.g xmin, ymin, zmin */
    geo_q_stop = 0xE003  # /* stop values, e.g xmax, ymax, zmax */
    geo_n_bins = 0xE004  # /* number of bins */
    geo_rotation = 0xE005  # [future] rotation of geometry
    geo_volume = 0xE006  # volume size in cm3 ... WARNING: may be a list in some future
    geo_zones = 0xE007  # single GEMCA zone, or list of zones
    geo_non_equidist_grid = 0xE008  # Array of non-equidistant z-grid. Tag only used if set.
    geo_units = 0xE009  # [Future]: ASCII string of ;-separated units along each dimension.
    geo_unit_ids = 0xE00A  # /* Unit IDs according to sh_units.h, one unit along each axis. */

    # Group 0xEF00 - 0xEFFF : Estimator
    filename_or_geotype = 0xEE00  # /* filename (bdo2019 format) or geometry name (bdo2016 format)*/
    estimator_number = 0xEE01  # /* Unique number for this estimator, if several files were saved, starting at 0 */
    number_of_pages = 0xEE02  # /* number of detectors / pages for this estimator */
    estimator_rescale_per_particle = 0xEE03  # estimator "per particle" rescaling, absent or set to 1 if no rescaling
    # note, that written data will *not* be multiplied with this value, it is up
    # to the BDO reader to multiply with this, if SHBDO_PAGE_NORMALIZE was set

    # /* Group 0xFFCC - 0xFFFF : Diagnostics, may be ignored by readers. */
    comment = 0xFFCC  # /* 0xFFCC-omment */
    debug = 0xFFCD  # /* 0xFFCD-ebug */
    error = 0xFFCE  # /* 0xFFCE-rror */


estimator_tags_to_save = (

)

page_tags_to_save = (
    SHBDOTagID.detector_type,
    SHBDOTagID.page_number,
    SHBDOTagID.page_normalized,
    SHBDOTagID.page_scale_factor,
    SHBDOTagID.page_offset,
    SHBDOTagID.page_unit_ids,

    SHBDOTagID.detector_unit,

    SHBDOTagID.page_diff_flag,
    SHBDOTagID.page_diff_type,
    SHBDOTagID.page_diff_start,
    SHBDOTagID.page_diff_stop,
    SHBDOTagID.page_diff_size,
    SHBDOTagID.page_diff_units,

    SHBDOTagID.page_filter_name,
    SHBDOTagID.page_filter_rules_no,
    SHBDOTagID.page_filter_e_min,
    SHBDOTagID.page_filter_emax,
)

# replace this dictionary with tuple estimator_tags_to_save
detector_name_from_bdotag = {
    SHBDOTagID.jpart0: 'projectile_code',
    SHBDOTagID.apro0: 'projectile_a',
    SHBDOTagID.zpro0: 'projectile_z',
    SHBDOTagID.beamx: 'projectile_position_x',
    SHBDOTagID.beamy: 'projectile_position_y',
    SHBDOTagID.beamz: 'projectile_position_z',
    SHBDOTagID.sigmax: 'projectile_sigma_x',
    SHBDOTagID.sigmay: 'projectile_sigma_y',
    SHBDOTagID.sigmat0: 'projectile_sigma_energy',
    SHBDOTagID.beamtheta: 'projectile_polar_angle',
    SHBDOTagID.beamphi: 'projectile_azimuth_angle',
    SHBDOTagID.beamdivx: 'projectile_divergence_x',
    SHBDOTagID.beamdivy: 'projectile_divergence_y',
    SHBDOTagID.beamdivk: 'projectile_divergence_k',
    SHBDOTagID.shversion: 'mc_code_version',
    SHBDOTagID.shbuilddate: 'mc_code_build_date',
    SHBDOTagID.filedate: 'filedate',
    SHBDOTagID.user: 'user',
    SHBDOTagID.host: 'host',
    SHBDOTagID.rt_nstat: 'number_of_primaries',
    SHBDOTagID.number_of_pages: 'page_count',
    SHBDOTagID.geo_unit_ids: 'geo_unit_ids',
    SHBDOTagID.geo_units: 'geo_units',
    SHBDOTagID.geometry_name: 'geo_name',
    SHBDOTagID.tmax0mev: 'Tmax_MeV',
    SHBDOTagID.tmax0amu: 'Tmax_MeV/amu',
    SHBDOTagID.tmax0nuc: 'Tmax_MeV/nucl'
}


class SHBDOUnitID(IntEnum):
    SH_SCORING_UNIT_UNKNOWN = -2  # /* e.g. for user defined scoring */
    SH_SCORING_UNIT_INVALID = -1  # /* N/A (not applicable), or undefined */

    SH_SCORING_UNIT_NONE = 0  # /* dimensionless */
    SH_SCORING_UNIT_AU = 1  # /* arbitrary units */
    SH_SCORING_UNIT_PCT = 2  # /* percent (%) */
    SH_SCORING_UNIT_PMIL = 3  # /* promill (%%) */
    SH_SCORING_UNIT_RELATIVE = 4  # /* dimensionless relative fraction */

    SH_SCORING_UNIT_CM = 10  # /* cm */
    SH_SCORING_UNIT_CM2 = 11  # /* cm^2 */
    SH_SCORING_UNIT_CM3 = 12  # /* cm^3 */
    SH_SCORING_UNIT_PCM = 13  # /* cm^-1 */
    SH_SCORING_UNIT_PCM2 = 14  # /* cm^-2 */
    SH_SCORING_UNIT_PCM3 = 15  # /* cm^-3 */

    SH_SCORING_UNIT_M = 16  # /* m */
    SH_SCORING_UNIT_M2 = 17  # /* m^2 */
    SH_SCORING_UNIT_M3 = 18  # /* m^3 */
    SH_SCORING_UNIT_PM = 19  # /* m^-1 */
    SH_SCORING_UNIT_PM2 = 20  # /* m^-2 */
    SH_SCORING_UNIT_PM3 = 21  # /* m^-3 */

    SH_SCORING_UNIT_GPCM3 = 22  # /* density g / cm^3 */
    SH_SCORING_UNIT_KGPM3 = 23  # /* density kg / m^3 */

    SH_SCORING_UNIT_KEVPUM = 30  # /* keV / um */
    SH_SCORING_UNIT_MEVPCM = 31  # /* MeV / cm */
    SH_SCORING_UNIT_MEVCM2PG = 32  # /* MeV * cm^2 / g */

    SH_SCORING_UNIT_MEVPG = 40  # /* MeV / g) */
    SH_SCORING_UNIT_GY = 41  # /* Gy (J/kg) */
    SH_SCORING_UNIT_GYRBE = 42  # /* Gy * RBE */
    SH_SCORING_UNIT_GYRE = 43  # /* Gy * RE (quenching) */
    SH_SCORING_UNIT_SV = 44  # /* Sv (J/kg) */
    SH_SCORING_UNIT_DOSERAD = 45  # /* Rad ( = 1 cGy)  */
    SH_SCORING_UNIT_DOSEREM = 46  # /* Rem ( = 1 cSv)  */

    SH_SCORING_UNIT_DEGREES = 50  # /* degrees */
    SH_SCORING_UNIT_RADIANS = 51  # /* radians */
    SH_SCORING_UNIT_SR = 52  # /* steradian sr = rad^2*/

    SH_SCORING_UNIT_COUNT = 60  # /* number # */

    SH_SCORING_UNIT_MEV = 70  # /* MeV */
    SH_SCORING_UNIT_MEVPNUC = 71  # /* MeV / nucleon */
    SH_SCORING_UNIT_MEVPAMU = 72  # /* MeV / amu */

    SH_SCORING_UNIT_NUCN = 80  # /* number of nucleons */
    SH_SCORING_UNIT_MEVPC2 = 81  # /* particle mass MeV / c^2 */
    SH_SCORING_UNIT_U = 82  # /* particle mass in terms of u */

    SH_SCORING_UNIT_MATID = 90  # /* material ID */
    SH_SCORING_UNIT_NZONE = 91  # /* zone number */


unit_name_from_unit_id = {
    SHBDOUnitID.SH_SCORING_UNIT_NONE: "",
    SHBDOUnitID.SH_SCORING_UNIT_AU: "a.u.",
    SHBDOUnitID.SH_SCORING_UNIT_PCT: "%",
    SHBDOUnitID.SH_SCORING_UNIT_PMIL: "%%",
    SHBDOUnitID.SH_SCORING_UNIT_RELATIVE: "rel.units",

    SHBDOUnitID.SH_SCORING_UNIT_CM: "cm",
    SHBDOUnitID.SH_SCORING_UNIT_CM2: "cm^2",
    SHBDOUnitID.SH_SCORING_UNIT_CM3: "cm^3",
    SHBDOUnitID.SH_SCORING_UNIT_PCM: "/cm",
    SHBDOUnitID.SH_SCORING_UNIT_PCM2: "/cm^2",
    SHBDOUnitID.SH_SCORING_UNIT_PCM3: "/cm^3",

    SHBDOUnitID.SH_SCORING_UNIT_M: "m",
    SHBDOUnitID.SH_SCORING_UNIT_M2: "m^2",
    SHBDOUnitID.SH_SCORING_UNIT_M3: "m^3",
    SHBDOUnitID.SH_SCORING_UNIT_PM: "/m",
    SHBDOUnitID.SH_SCORING_UNIT_PM2: "/m^2",
    SHBDOUnitID.SH_SCORING_UNIT_PM3: "/m^3",

    SHBDOUnitID.SH_SCORING_UNIT_GPCM3: "g/cm^3",
    SHBDOUnitID.SH_SCORING_UNIT_KGPM3: "kg/m^3",

    SHBDOUnitID.SH_SCORING_UNIT_KEVPUM: "keV/um",
    SHBDOUnitID.SH_SCORING_UNIT_MEVPCM: "MeV/cm",
    SHBDOUnitID.SH_SCORING_UNIT_MEVCM2PG: "MeV cm^2/g",

    SHBDOUnitID.SH_SCORING_UNIT_MEVPG: "MeV/g",
    SHBDOUnitID.SH_SCORING_UNIT_GY: "Gy",
    SHBDOUnitID.SH_SCORING_UNIT_GYRBE: "Gy(RBE)",  # /* Probably there are new ICRU rules here */
    SHBDOUnitID.SH_SCORING_UNIT_GYRE: "Gy(RE)",
    SHBDOUnitID.SH_SCORING_UNIT_SV: "Sv",
    SHBDOUnitID.SH_SCORING_UNIT_DOSERAD: "Rad",
    SHBDOUnitID.SH_SCORING_UNIT_DOSEREM: "Rem",

    SHBDOUnitID.SH_SCORING_UNIT_DEGREES: "deg",
    SHBDOUnitID.SH_SCORING_UNIT_RADIANS: "rad",
    SHBDOUnitID.SH_SCORING_UNIT_SR: "sr",

    SHBDOUnitID.SH_SCORING_UNIT_COUNT: "#",

    SHBDOUnitID.SH_SCORING_UNIT_MEV: "MeV",
    SHBDOUnitID.SH_SCORING_UNIT_MEVPNUC: "MeV/nucleon",
    SHBDOUnitID.SH_SCORING_UNIT_MEVPAMU: "MeV/amu",

    SHBDOUnitID.SH_SCORING_UNIT_NUCN: "",
    SHBDOUnitID.SH_SCORING_UNIT_MEVPC2: "MeV/c^2",
    SHBDOUnitID.SH_SCORING_UNIT_U: "u",

    SHBDOUnitID.SH_SCORING_UNIT_MATID: "",
    SHBDOUnitID.SH_SCORING_UNIT_NZONE: "",
}



from enum import IntEnum
import logging

import numpy as np

from pymchelper.readers.common import ReaderFactory
from pymchelper.readers.shieldhit.reader_base import read_next_token
from pymchelper.readers.shieldhit.reader_bdo2016 import SHReaderBDO2016
from pymchelper.readers.shieldhit.reader_bdo2019 import SHReaderBDO2019
from pymchelper.readers.shieldhit.reader_bin2010 import SHReaderBin2010
from pymchelper.readers.shieldhit.binary_spec import SHBDOTagID

logger = logging.getLogger(__name__)


def file_has_sh_magic_number(filename):
    """
    BDO binary files, introduced in 2016 (BDO2016 and BDO2019 formats) starts with 6 magic bytes xSH12A
    :param filename: Binary file filename
    :return: True if binary file starts with SH magic number
    """
    sh_bdo_magic_number = b'xSH12A'
    has_bdo_magic_number = False
    with open(filename, "rb") as f:
        d1 = np.dtype([('magic', 'S6')])  # TODO add a check if file has less than 6 bytes or is empty
        x = np.fromfile(f, dtype=d1, count=1)
        if x:
            # compare first 6 bytes with reference string
            has_bdo_magic_number = (sh_bdo_magic_number == x['magic'][0])

    logger.debug("File {:s} has magic number: {:s}".format(filename, str(has_bdo_magic_number)))
    return has_bdo_magic_number


def extract_sh_ver(filename):
    """
    BDO binary files, introduced in 2016 (BDO2016 and BDO2019 formats) contain information about SH VER
    :param filename: Binary file filename
    :return: SH12 version (as a string, i.e. 0.7) or None if version information was not found in the file
    """

    ver = None
    with open(filename, "rb") as f:
        d1 = np.dtype([('magic', 'S6'),
                       ('end', 'S2'),
                       ('vstr', 'S16')])  # TODO add a check if file has less than 6 bytes or is empty
        x = np.fromfile(f, dtype=d1, count=1)
        logger.debug("File {:s}, raw version info {:s}".format(filename, str(x['vstr'][0])))
        try:
            ver = x['vstr'][0].decode('ASCII')
        except UnicodeDecodeError:
            ver = None

    logger.debug("File {:s}, SH12A version: {:s}".format(filename, str(ver)))
    return ver


class SHFileFormatId(IntEnum):
    """
    SHIELD-HIT12A file format ids, as described in sh_file_format.h file
    """
    bin2010 = 0   # Old binary format from 2010, the first version David wrote back then
    bdo2016 = 1   # year 2016 .bdo file format, now with proper tags, used from SH 0.6.0
    bdo2019 = 2   # year 2019 .bdo style, introduced June 2019
    ascii = 3     # raw text format
    csv = 4       # comma separated file format


def read_token(filename, token_id):
    """
    TODO
    :param filename:
    :param token_id:
    :return:
    """
    with open(filename, "rb") as f:

        # skip ASCII header
        d1 = np.dtype([('magic', 'S6'), ('endiannes', 'S2'), ('vstr', 'S16')])
        np.fromfile(f, dtype=d1, count=1)

        # read tokens from rest of the file
        while f:
            token = read_next_token(f)
            if token is None:
                break
            pl_id, _pl_type, _pl_len, _pl = token
            if pl_id == token_id:

                logger.debug("Read token {:s} (0x{:02x}) value {} type {:s} length {:d}".format(
                    SHBDOTagID(pl_id).name,
                    pl_id,
                    _pl,
                    _pl_type.decode('ASCII'),
                    _pl_len
                ))

                pl = [None] * _pl_len

                # decode all strings (currently there will never be more than one per token)
                if 'S' in _pl_type.decode('ASCII'):
                    for i, _j in enumerate(_pl):
                        pl[i] = _pl[i].decode('ASCII').strip()
                else:
                    pl = _pl

                if len(pl) == 1:
                    pl = pl[0]

                return pl
        return None


class SHReaderFactory(ReaderFactory):
    def get_reader(self):
        """
        Inspect binary file and return appropriate reader object
        :return:
        """
        reader = None

        # TODO add ZIP file unpacking

        # magic number was introduced together with first token-based BDO file format (BDO2016)
        # presence of magic number means we could have BDO2016 or BDO2019 format
        if file_has_sh_magic_number(self.filename):
            reader = SHReaderBDO2019

            # format tag specifying binary standard was introduced in SH12A v0.7.4-dev on  07.06.2019 (commit 6eddf98)
            file_format = read_token(self.filename, SHBDOTagID.format)
            if file_format:
                logger.debug("File format: {} {:s}".format(file_format, SHFileFormatId(file_format).name))
                if file_format == SHFileFormatId.bdo2019:
                    reader = SHReaderBDO2019
                elif file_format == SHFileFormatId.bdo2016:
                    reader = SHReaderBDO2016
                else:
                    print("What shall we do ?")
            else:
                # in case format tag is missing we default to BDO2016 format
                # this mean we cannot read BDO2019 files generated with SH12A built before 07.06.2019
                logger.info("File format information missing (token)")
                reader = SHReaderBDO2016
        else:
            # lack of magic number means we expect Fortran-style binary format (BIN2010)
            reader = SHReaderBin2010

        # ver_short = extract_sh_ver(self.filename)
        # logger.info("Short version: {:s}".format(str(ver_short)))
        # ver_long = read_token(self.filename, SHBDOTagID.shversion)
        # logger.info("Long version: {:s}".format(str(ver_long)))

        return reader


class SHReaderASCII:
    """
    Reads plain text files with data saved by binary-to-ascii converter.
    """

    def __init__(self, filename):
        self.filename = filename

    def read_header(self, detector):
        raise NotImplementedError

    def read_payload(self, detector):
        raise NotImplementedError

    def read(self, detector):
        self.read_header(detector)
        self.read_payload(detector)



import logging
import numpy as np

logger = logging.getLogger(__name__)


class HdfWriter:
    """
    Supports writing HDF file format.
    HDF is designed to store large amounts of data organized in convenient way.
    One HDF file can handle many single- or multi-dimensional tables.

    """
    def __init__(self, filename, options):
        self.filename = filename
        if not self.filename.endswith(".h5"):
            self.filename += ".h5"

    def write(self, estimator):
        if len(estimator.pages) > 1:
            print("Conversion of data with multiple pages not supported yet")
            return False

        try:
            import h5py
        except ImportError as e:
            logger.error("Generating HDF5 files not available on your platform (please install h5py).")
            raise e

        with h5py.File(self.filename, 'w') as f:

            # change units for LET from MeV/cm to keV/um if necessary
            # a copy of data table is made here
            # from pymchelper.shieldhit.detector.detector_type import SHDetType
            # if estimator.dettyp in (SHDetType.dlet, SHDetType.dletg, SHDetType.tlet, SHDetType.tletg):
            #     data = estimator.data * np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um
            #     if not np.all(np.isnan(estimator.error_raw)) and np.any(estimator.error_raw):
            #         error = estimator.error * np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um
            # else:
            #     data = estimator.data
            #     error = estimator.error
            # TODO move to reader

            page = estimator.pages[0]
            data = page.data
            error = page.error

            # save data
            dset = f.create_dataset("data", data=data, compression="gzip", compression_opts=9)

            # save error (if present)
            if not np.all(np.isnan(page.error_raw)) and np.any(page.error_raw):
                f.create_dataset("error", data=error, compression="gzip", compression_opts=9)

            # save metadata
            dset.attrs['name'] = page.name
            dset.attrs['unit'] = page.unit
            dset.attrs['nstat'] = estimator.number_of_primaries
            dset.attrs['counter'] = estimator.file_counter
            dset.attrs['xaxis_n'] = estimator.x.n
            dset.attrs['xaxis_min'] = estimator.x.min_val
            dset.attrs['xaxis_max'] = estimator.x.max_val
            dset.attrs['xaxis_name'] = estimator.x.name
            dset.attrs['xaxis_unit'] = estimator.x.unit
            dset.attrs['yaxis_n'] = estimator.y.n
            dset.attrs['yaxis_min'] = estimator.y.min_val
            dset.attrs['yaxis_max'] = estimator.y.max_val
            dset.attrs['yaxis_name'] = estimator.y.name
            dset.attrs['yaxis_unit'] = estimator.y.unit
            dset.attrs['zaxis_n'] = estimator.z.n
            dset.attrs['zaxis_min'] = estimator.z.min_val
            dset.attrs['zaxis_max'] = estimator.z.max_val
            dset.attrs['zaxis_name'] = estimator.z.name
            dset.attrs['zaxis_unit'] = estimator.z.unit

        return 0



import logging
import os
from collections import defaultdict
from glob import glob

import numpy as np

from pymchelper.estimator import ErrorEstimate, Estimator, average_with_nan
from pymchelper.readers.fluka import FlukaReader, FlukaReaderFactory
from pymchelper.readers.shieldhit.general import SHReaderFactory
from pymchelper.readers.shieldhit.reader_base import SHReader
from pymchelper.writers.common import Converters

logger = logging.getLogger(__name__)


def guess_reader(filename):
    """
    Guess a reader based on file contents or extensions.
    In some cases (i.e. binary SH12A files) access to file contents is needed.
    :param filename:
    :return: Instantiated reader object
    """
    fluka_reader = FlukaReaderFactory(filename).get_reader()
    if fluka_reader:
        reader = fluka_reader(filename)
    else:
        sh_reader = SHReaderFactory(filename).get_reader()
        if sh_reader:
            reader = sh_reader(filename)
    return reader


def guess_corename(filename):
    """
    Guess a reader based on file contents or extensions.
    In some cases (i.e. binary SH12A files) access to file contents is needed.
    :param filename:
    :return: the corename of the file (i.e. the basename without the running number for averaging)
    """
    corename = FlukaReader(filename).corename
    if corename is None:
        corename = SHReader(filename).corename
    return corename


def fromfile(filename):
    """Read estimator data from a binary file ```filename```"""

    reader = guess_reader(filename)
    if reader is None:
        raise Exception("File format not compatible", filename)
    estimator = Estimator()
    estimator.file_counter = 1
    if not reader.read(estimator):  # some problems occurred during read
        estimator = None
    return estimator


def fromfilelist(input_file_list, error=ErrorEstimate.stderr, nan: bool = True):
    """
    Reads all files from a given list, and returns a list of averaged estimators.

    :param input_file_list: list of files to be read
    :param error: error estimation, see class ErrorEstimate class in pymchelper.estimator
    :param nan: if True, NaN (not a number) are excluded when averaging data.
    :return: list of estimators
    """
    if not isinstance(input_file_list, list):  # probably a string instead of list
        input_file_list = [input_file_list]

    if nan:
        estimator_list = [fromfile(filename) for filename in input_file_list]
        result = average_with_nan(estimator_list, error)
        if not result:  # TODO check here !
            return None
    elif len(input_file_list) == 1:
        result = fromfile(input_file_list[0])
        if not result:
            return None
    else:
        result = fromfile(input_file_list[0])
        if not result:
            return None

        # allocate memory for accumulator in standard deviation calculation
        # not needed if user requested not to include errors
        if error != ErrorEstimate.none:
            for page in result.pages:
                page.error_raw = np.zeros_like(page.data_raw)

        # loop over all files with n running from 2
        for n, filename in enumerate(input_file_list[1:], start=2):
            current_estimator = fromfile(filename)  # x

            # Running variance algorithm based on algorithm by B. P. Welford,
            # presented in Donald Knuth's Art of Computer Programming, Vol 2, page 232, 3rd edition.
            # Can be found here: http://www.johndcook.com/blog/standard_deviation/
            # and https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Online_algorithm
            delta = [
                current_page.data_raw - result_page.data_raw
                for current_page, result_page in zip(current_estimator.pages, result.pages)
            ]  # delta = x - mean
            for page, delta_item in zip(result.pages, delta):
                page.data_raw += delta_item / np.float64(n)

            if error != ErrorEstimate.none:
                for page, delta_item, current_page in zip(result.pages, delta, current_estimator.pages):
                    page.error_raw += delta_item * (current_page.data_raw - page.data_raw)  # M2 += delta * (x - mean)

        # unbiased sample variance is stored in `__M2 / (n - 1)`
        # unbiased sample standard deviation in classical algorithm is calculated as (sqrt(1/(n-1)sum(x-<x>)**2)
        # here it is calculated as square root of unbiased sample variance:
        if len(input_file_list) > 1 and error != ErrorEstimate.none:
            for page in result.pages:
                page.error_raw = np.sqrt(page.error_raw / (len(input_file_list) - 1.0))

        # if user requested standard error then we calculate it as:
        # S = stderr = stddev / sqrt(N), or in other words,
        # S = s/sqrt(N) where S is the corrected standard deviation of the mean.
        if len(input_file_list) > 1 and error == ErrorEstimate.stderr:
            for page in result.pages:
                page.error_raw /= np.sqrt(len(input_file_list))  # np.sqrt() always returns np.float64

    result.file_counter = len(input_file_list)
    core_names_dict = group_input_files(input_file_list)
    if len(core_names_dict) == 1:
        result.file_corename = list(core_names_dict)[0]

    return result


def frompattern(pattern, error=ErrorEstimate.stderr, nan=True):
    """
    Reads all files matching pattern, e.g.: 'foobar_*.bdo', and returns a list of averaged estimators.

    :param pattern: pattern to be matched for reading.
    :param error: error estimation, see class ErrorEstimate class in pymchelper.estimator
    :param nan: if True, NaN (not a number) are excluded when averaing data.
    :return: a list of estimators, or an empty list if no files were found.
    """

    try:
        list_of_matching_files = glob(pattern)
    except TypeError as e:  # noqa: F841
        list_of_matching_files = pattern

    core_names_dict = group_input_files(list_of_matching_files)

    result = [fromfilelist(filelist, error, nan) for _, filelist in core_names_dict.items()]
    return result


def convertfromlist(filelist, error, nan, outputdir, converter_name, options, outputfile=None):
    """

    :param filelist:
    :param error: error estimation, see class ErrorEstimate class in pymchelper.estimator
    :param nan: if True, NaN (not a number) are excluded when averaing data.
    :param outputdir:
    :param converter_name:
    :param options:
    :param outputfile:
    :return:
    """
    estimator = fromfilelist(filelist, error, nan)
    if not estimator:
        return None
    if outputfile is not None:
        output_path = outputfile
    elif outputdir is None:
        output_path = estimator.file_corename
    else:
        output_path = os.path.join(outputdir, estimator.file_corename)
    status = tofile(estimator, output_path, converter_name, options)
    return status


def convertfrompattern(pattern,
                       outputdir,
                       converter_name,
                       options,
                       error=ErrorEstimate.stderr,
                       nan: bool = True):
    """

    :param pattern:
    :param outputdir:
    :param converter_name:
    :param options:
    :param error: error estimation, see class ErrorEstimate class in pymchelper.estimator
    :param nan: if True, NaN (not a number) are excluded when averaging data.
    :return:
    """
    list_of_matching_files = glob(pattern)

    core_names_dict = group_input_files(list_of_matching_files)

    status = []
    for _, filelist in core_names_dict.items():
        status.append(convertfromlist(filelist, error, nan, outputdir, converter_name, options))
    return max(status)


def tofile(estimator, filename, converter_name, options):
    """
    Save a estimator data to a ``filename`` using converter defined by ``converter_name``
    :param estimator:
    :param filename:
    :param converter_name:
    :param options:
    :return:
    """
    writer_cls = Converters.fromname(converter_name)
    writer = writer_cls(filename, options)
    logger.debug(f"File corename : {filename}")
    status = writer.write(estimator)
    return status


def group_input_files(input_file_list):
    """
    Takes set of input file names, belonging to possibly different estimators.
    Input files are grouped according to the estimators and for each group
    merging is performed, as in @merge_list method.
    Output file name is automatically generated.
    :param input_file_list: list of input files
    :return: core_names_dict
    """
    core_names_dict = defaultdict(list)
    # keys - core_name, value - list of full paths to corresponding files

    # loop over input list of file paths
    for filepath in input_file_list:
        core_name = guess_corename(filepath)
        core_names_dict[core_name].append(filepath)

    return core_names_dict
